"""
This file defines DatapointReference and DatasetReference

A dataset reference consists of many data point references. Its job is to make
clear what data files are in the dataset and how they should be read into the
memory (when the dataset is used). A dataset reference can therefore be used
to pass data into a workload, and the reference can be stored as a file afterward
to make the workload reproducible.
"""


from datetime import datetime
from cvpl_tools.array_key_dict import ArrayKeyDict
import cvpl_tools.fs as fs
import numpy as np
from dataclasses import dataclass


@dataclass
class DatapointReference:
    """
    Represents a pointer to location of data.
    The data is either an image or a chunk (data point) in the dataset, and one such
    datapoint can always be represented as a single numpy array.
    """
    data_ref: str | list

    def __init__(self, data_ref: str | list):
        """
        Create a DatapointReference object
        Args:
            data_ref: A list of strings OR a single string containing link to one or more files referred to by
                this datapoint reference
        """
        assert isinstance(data_ref, (str, list)), 'ERROR: Datapoint reference should be either a string or a list!'
        self.data_ref = data_ref

    def __str__(self):
        """
        Returns:
            A readable string representation of the link(s)
        """
        return str(self.data_ref)

    def ref(self):
        """
        Returns:
            A list of strings OR a single string containing link to one or more files referred to by
            this datapoint reference
        """
        return self.data_ref

    def has_multiple_images(self):
        """
        Returns:
            True if the datapoint is described by multiple images (in which case there is a list of
            links pointing to these images)
        """
        return isinstance(self.data_ref, list)

    def read_as_np(self, read_setting: fs.ImReadSetting) -> np.ndarray:
        """
        Reads the datapoint as a numpy array
        Args:
            read_setting: An cvpl_tools.fs.ImReadSetting object that describes how an image file is
            readed into memory into a numpy array
        Returns:
            The readed numpy array
        """
        return fs.ImIO.read_single_image(read_setting, self.data_ref)[0]


@dataclass
class DatasetReference:
    """
    This class' instance represents a dataset, and whose attributes include information on what data is in the dataset
    and when and how this reference is created.

    The datapoints in a dataset should be homogeneous - meaning they can be processed in the same way, and images
    should be stored in a separate dataset as its corresponding labels
    """
    datapoint_refs: ArrayKeyDict[str, DatapointReference]
    im_read_setting: fs.ImReadSetting
    creation_date: datetime
    creation_info: str  # a description of how this dataset is created
    name: str  # name of the dataset reference

    @staticmethod
    def new(datapoint_refs: ArrayKeyDict[str, DatapointReference],
            dataset_name: str,
            creation_info: str = '',
            im_read_setting: fs.ImReadSetting = None):
        """
        The recommended interface to create a DatasetReference object.
        Args:
            datapoint_refs: Mapping datapoint ids (imids) to DatapointReference objects
            dataset_name: Name of the dataset
            creation_info: A description of how the dataset is created; for better reproducibility archiving
            im_read_setting: A setting object that will be used to read images into memory (if needed)

        Returns:

        """
        return DatasetReference(
            datapoint_refs=datapoint_refs,
            creation_date=datetime.now(),
            creation_info=creation_info,
            name=dataset_name,
            im_read_setting=im_read_setting
        )

    @staticmethod
    def empty():
        """
        Create an empty DatasetReference object
        Returns:
            The created DatasetReference object
        """
        return DatasetReference.new(ArrayKeyDict(),
                                'Empty Dataset',
                                'This dataset is created as an empty dataset.',
                                None)


def test():
    in1 = 'C:/path/to/file1'
    r1 = DatapointReference(in1)
    in2 = ['C:/path/to/file2', 'C:/path/to/file3']
    r2 = DatapointReference(in2)
    assert str(r1) == 'C:/path/to/file1'
    assert len(r2.ref()) == 2 and r2.ref()[0] == 'C:/path/to/file2'



"""
Segmentation and post-processing.

This file is for methods generating (dask) single class segmentation masks of binary or ordinal types, the latter of
which is a 0-N segmentation of N objects of the same class in an image.

Methods in this file should be run quickly and whose performance can be compared against each other to manual
segmentations over a dataset.

The input to these methods are either input 3d single-channel image of type np.float32, or input image paired with
a deep learning segmentation algorithm. The output may be cell count #, binary mask
(np.uint8) or ordinal mask (np.int32).

Conceptually, we define the following input/output types:
IN - Input Image (np.float32) between 0 and 1, this is the brightness dask image as input
BS - Binary Segmentation (3d, np.uint8), this is the binary mask single class segmentation
OS - Ordinal Segmentation (3d, np.int32), this is the 0-N where contour 1-N each denotes an object; also single class
LC - List of Centroids, this contains a list of centroids for each block in the original image
CC - Cell Count Map (3d, np.float32), a cell count number (estimate, can be float) for each block
CD - Cell Density Map (3d, np.float32), this is a down sampled map of the cell density of the brain
ATLAS_MAP - A function that maps from the pixel location within brain to brain regions
ATLAS_CD - Atlas, this summarizes cell density about the brain grouped by region of the brain

Each type above may have associated parameters of them that is not changed.
And we work with the following types of methods as steps to obtain final segmentations from the input mask:
preprocess: IN -> IN
e.g. gaussian_blur

predict_bs: IN -> BS
e.g. cellseg3d_predict, simple_threshold

predict_lc: IN -> LC
e.g. blob_dog

predict_cc: IN -> CC
e.g. scaled_sum_intensity

binary_to_inst: BS -> OS
e.g. direct_bs_to_os, watershed

binary_and_centroids_to_inst: (BS, LC) -> OS
e.g. in_contour_voronoi

os_to_lc: OS -> LC
e.g. direct_os_to_centroids

count_from_lc: LC -> CC
e.g. count_lc_ncentroid, count_edge_penalized_lc_ncentroid

count_from_os: OS -> CC
e.g. count_os_ncontour, count_edge_penalized_os_ncontour, count_os_byvolume

count_to_atlas_cell_density: CC -> ATLAS_CD
e.g. average_by_region

Each method is an object implementing the SegStage interface that has the following methods:
- input_type() -> ty
- output_type() -> ty
- forward(*args) -> out, this is the forward stage
- interpretable_napari(viewer, *args) -> None, this adds the appropriate human-interpretable debugging-purpose outputs to viewer

About the dask version of the last 2 functions: These are for parallel processing of the images; they need not be
interpretable for large images but the interpretable_napari function should handle small sized dask images for
debugging purpose.
"""

import abc
import enum
import logging
from typing import Callable, Any, Sequence, final, Iterator

import cvpl_tools.fs as fs
import cvpl_tools.ome_zarr.io as cvpl_ome_zarr_io
import cvpl_tools.im.algorithms as algorithms
import cvpl_tools.im.dask_algorithms as dask_algorithms
from cvpl_tools.im.dask_algorithms import NDBlock
import dask.array as da
import napari
import numpy as np
import skimage
from scipy.ndimage import (
    gaussian_filter as gaussian_filter,
    label as instance_label,
    find_objects as find_objects
)

_logger = logging.getLogger('SEG_PROCESS')


# ------------------------------------Helper Functions---------------------------------------


def direct_map_block(forward: Callable, im: da.Array):
    def map_fn(block, block_info=None):
        return forward(block)

    return im.map_blocks(map_fn)


def lc_interpretable_napari(layer_name: str,
                            lc: np.ndarray,
                            viewer: napari.Viewer,
                            ndim: int,
                            extra_features: Sequence):
    """This function is used to display feature points for LC-typed output

    Args:
        layer_name: displayed name of the layer
        lc: The list of features, each row of length (ndim + nextra)
        viewer: Napari viewer to add points to
        ndim: dimension of the image
        extra_features: extra features to be displayed as text
    """
    # reference: https://napari.org/stable/gallery/add_points_with_features.html
    nextra = len(extra_features)
    assert isinstance(lc, np.ndarray), 'lc should be of type np.ndarray!'
    assert lc.ndim == 2, (f'Wrong dimension for list of centroids, expected ndim=2, but got lc={lc} and '
                          f'lc.shape={lc.shape}')
    assert lc.shape[1] == nextra + ndim, (f'Wrong number of features for list of centroids, expected length along '
                                          f'first dimension to be nextra + ndim={nextra + ndim} but got lc={lc} and '
                                          f'lc.shape={lc.shape}')

    features = {
        extra_features[i]: lc[:, ndim + i] for i in range(nextra)
    }

    strings = [extra_features[i] + '={' + extra_features[i] + ':.2f}' for i in range(nextra)]
    text_parameters = {
        'string': '\n'.join(strings),
        'size': 9,
        'color': 'green',
        'anchor': 'center',
    }
    viewer.add_points(lc[:, :ndim],
                      size=1.,
                      ndim=ndim,
                      name=layer_name,
                      features=features,
                      text=text_parameters)


# ---------------------------------------Interfaces------------------------------------------


logger = logging.getLogger('SEG_PROCESSES')


class DatType(enum.Enum):
    IN = 0
    BS = 1
    OS = 2
    LC = 3
    CC = 4
    CD = 5
    ATLAS_MAP = 6
    ATLAS_CD = 7
    OTHER = 8  # tuple of two or more of the above, or etc.


class SegProcess(abc.ABC):
    def __init__(self, input_type: DatType, output_type: DatType):
        self._input_type = input_type
        self._output_type = output_type
        self.tmpdir: fs.TmpDirectory | None = None

    def input_type(self) -> DatType:
        return self._input_type

    def output_type(self) -> DatType:
        return self._output_type

    def input_type_str(self) -> str:
        return self._input_type.name

    def output_type_str(self) -> str:
        return self._output_type.name

    def set_tmpdir(self, tmpdir: fs.TmpDirectory):
        """Use this to cache output results

        For output images that involves interpretation with napari, this is necessary for a smooth display of the
        results.

        Args:
            tmpdir: Any fs.TmpDirectory object that allows many temporary objects to be written as cache in the
                directory
        """
        assert isinstance(tmpdir, fs.TmpDirectory), f'Expected type cvpl_tools.fs.TempDirectory, got {type(tmpdir)}'
        self.tmpdir = tmpdir

    def cache_im(self, im: np.ndarray | da.Array) -> np.ndarray | da.Array:
        if isinstance(im, da.Array):
            assert self.tmpdir is not None, 'Please assign a tmp_prefix before caching Napari image!'
            tmp_prefix = self.tmpdir.assign_tmpdir()
            im = cvpl_ome_zarr_io.cache_image(im, tmp_prefix)
        return im

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        raise NotImplementedError("SegProcess base class does not implement forward()!")

    @abc.abstractmethod
    def interpretable_napari(self, *args, **kwargs):
        """By default, we directly try to add the output to Napari

        This is the simplest way to interpret the output, but will not work for all data types. Subclasses should
        disregard this method and write their own if any more complicated interpretation is needed.

        Args:
            viewer: Napari viewer object
            *args: args to be passed to forward()
            **kwargs: keyword args to be passed to forward()
        """
        raise NotImplementedError("SegProcess base class does not implement forward()!")


class BlockToBlockProcess(SegProcess):
    def __init__(self, input_type: DatType, output_type: DatType, is_label=False):
        super().__init__(input_type, output_type)
        self.is_label = is_label

    @final
    def forward(self, im: np.ndarray | da.Array) \
            -> np.ndarray | da.Array:
        if isinstance(im, np.ndarray):
            return self.np_forward(im)
        elif isinstance(im, da.Array):
            return self.dask_forward(im)
        else:
            raise TypeError(f'Invalid im type: {type(im)}')

    @abc.abstractmethod
    def np_forward(self, im: np.ndarray) -> np.ndarray:
        raise NotImplementedError('Subclass of BlockToBlockProcess should implement np_forward!')

    def dask_forward(self, im: da.Array) -> da.Array:
        return direct_map_block(self.np_forward, im)

    def interpretable_napari(self, viewer: napari.Viewer, im: np.ndarray | da.Array):
        fn = viewer.add_labels if self.is_label else viewer.add_image
        fn(self.cache_im(self.forward(im)), name='interpretation')


def forward_sequential_processes(processes: Sequence[SegProcess], arg: Any):
    """Run a sequence of processes as one

    Args:
        processes: The list of SegProcess objects to be chained to one
        arg: The argument to the first SegProcess object

    Returns:
        The result returned after a chain of operations
    """
    cur_out = arg
    for i in range(len(processes)):
        cur_out = processes[i].forward(cur_out)
    return cur_out


def interpretable_napari_sequential_processes(viewer: napari.Viewer,
                                              processes: Sequence[SegProcess],
                                              arg: Any,
                                              add_flags: Sequence[bool] | None = None):
    """Run a sequence of processes, and interpret these by adding them into napari as images

    Args:
        viewer: Napari viewer
        processes: The processes to be interpreted
        arg: The argument to the first process
        add_flags: default to all True; a list of flags indicating which processes in the list should be interpreted
    """
    if add_flags is None:
        add_flags = [True] * len(processes)
    last_true = -1
    for i in range(len(processes)):
        if add_flags[i]:
            last_true = i
    if last_true == -1:
        return

    cur_out = arg
    for i in range(last_true + 1):
        if i != last_true:
            cur_out = processes[i].forward(cur_out)
        processes[i].interpretable_napari(cur_out, viewer=viewer)


# ---------------------------------------Preprocess------------------------------------------


class GaussianBlur(BlockToBlockProcess):
    def __init__(self, sigma: float):
        super().__init__(DatType.IN, DatType.IN, is_label=False)
        self.sigma = sigma

    def np_forward(self, im: np.ndarray[np.float32]) -> np.ndarray[np.float32]:
        return gaussian_filter(im, sigma=self.sigma)


# -------------------------------------Predict Binary----------------------------------------


class BSPredictor(BlockToBlockProcess):
    def __init__(self, pred_fn: Callable):
        super().__init__(DatType.IN, DatType.BS, is_label=True)
        self.pred_fn = pred_fn

    def np_forward(self, im: np.ndarray[np.float32]) -> np.ndarray[np.uint8]:
        return self.pred_fn(im)


class SimpleThreshold(BlockToBlockProcess):
    def __init__(self, threshold: float):
        super().__init__(DatType.IN, DatType.BS, is_label=True)
        self.threshold = threshold

    def np_forward(self, im: np.ndarray[np.float32]) -> np.ndarray[np.uint8]:
        return (im > self.threshold).astype(np.uint8)


# --------------------------------Predict List of Centroids-----------------------------------


class BlobDog(SegProcess):
    def __init__(self, max_sigma=2, threshold: float = 0.1, reduce=False):
        super().__init__(DatType.IN, DatType.LC)
        self.max_sigma = max_sigma
        self.threshold = threshold
        self.reduce = reduce

    def np_features(self, block: np.ndarray[np.float32], block_info=None) -> np.ndarray[np.float32]:
        if block_info is not None:
            slices = block_info[0]['array-location']
            lc = skimage.feature.blob_dog(np.array(block * 255, dtype=np.uint8),
                                          max_sigma=self.max_sigma,
                                          threshold=self.threshold).astype(np.float32)  # N * (ndim + 1) ndarray
            start_pos = np.array([slices[i].start for i in range(len(slices))], dtype=np.float32)
            lc[:, :block.ndim] += start_pos[None, :]
            return lc
        else:
            return block

    def feature_forward(self, im: np.ndarray[np.float32] | da.Array) -> NDBlock[np.float32]:
        return NDBlock.map_ndblocks([NDBlock(im)], self.np_features, out_dtype=np.float32)

    def forward(self, im: np.ndarray[np.float32] | da.Array) -> NDBlock:
        ndblock = self.feature_forward(im)
        ndblock = ndblock.select_columns(slice(im.ndim))
        if self.reduce:
            ndblock = ndblock.reduce(force_numpy=False)
        return ndblock

    def interpretable_napari(self, viewer: napari.Viewer, im: np.ndarray[np.float32]):
        blobdog = self.feature_forward(im).reduce(force_numpy=True)
        lc_interpretable_napari('blobdog_centroids', blobdog, viewer, im.ndim, ['sigma'])


# -------------------------------Direct Cell Count Prediction----------------------------------


class ScaledSumIntensity(SegProcess):
    def __init__(self, scale: float = .008, min_thres: float = 0., reduce: bool = True, spatial_box_width: int = 8):
        """Summing up the intensity and scale it to obtain number of cells, directly

        Args:
            scale: Scale the sum of intensity by this to get number of cells
            min_thres: Intensity below this threshold is excluded (set to 0 before summing)
        """
        assert scale >= 0, f'{scale}'
        assert 0. <= min_thres <= 1., f'{min_thres}'
        super().__init__(DatType.IN, DatType.CC)
        self.scale = scale
        self.min_thres = min_thres
        self.reduce = reduce
        self.spatial_box_width = spatial_box_width

    def np_features(self, block: np.ndarray[np.float32], block_info=None, spatial_box_width=None) \
            -> np.ndarray[np.float32]:
        if block_info is not None:
            slices = block_info[0]['array-location']
            if spatial_box_width is not None:
                padded_block = algorithms.pad_to_multiple(block, spatial_box_width)
            else:
                padded_block = block
            masked = padded_block * (padded_block > self.min_thres)
            if slices is not None:
                startpoint = np.array([slices[i].start for i in range(len(slices))],
                                      dtype=np.float32)
            else:
                startpoint = np.zeros((block.ndim,), dtype=np.float32)

            if spatial_box_width is not None:
                subblock_shape = (spatial_box_width,) * block.ndim
                masked = algorithms.np_map_block(masked, block_sz=subblock_shape)
                masked: np.ndarray = masked.sum(axis=tuple(range(block.ndim, block.ndim * 2)))

                features = np.zeros((masked.size, block.ndim + 1), dtype=np.float32)
                inds = np.array(np.indices(masked.shape, dtype=np.float32))
                transpose_axes = tuple(range(1, block.ndim + 1)) + (0,)
                inds = inds.transpose(transpose_axes).reshape(-1, block.ndim)
                features[:, -1] = masked.flatten() * self.scale
                features[:, :-1] = startpoint[None, :] + (inds + .5) * np.array(subblock_shape, dtype=np.float32)
            else:
                features = np.zeros((1, block.ndim + 1), dtype=np.float32)
                features[0, -1] = masked.sum() * self.scale
                features[:, :-1] = startpoint[None, :] + np.array(block.shape, dtype=np.float32) / 2
            return features
        else:
            return block

    def feature_forward(self, im: np.ndarray[np.float32] | da.Array, spatial_block_width: int = None) \
            -> NDBlock[np.float32]:
        def map_fn(b, block_info):
            return self.np_features(b, block_info, spatial_block_width)

        return NDBlock.map_ndblocks([NDBlock(im)], map_fn, out_dtype=np.float32)

    def forward(self, im: np.ndarray | da.Array) \
            -> NDBlock | np.ndarray:
        ndblock = self.feature_forward(im)
        ndblock = ndblock.select_columns([-1])
        if self.reduce:
            ndblock = ndblock.reduce(force_numpy=False)
        return ndblock

    def interpretable_napari(self, viewer: napari.Viewer, im: np.ndarray[np.float32] | da.Array):
        # see what has been completely masked off
        mask = im > self.min_thres
        viewer.add_image(self.cache_im(mask), name='spatial_vis_ncell_sum_intensity')

        # see for each spatial block, how many cells are counted within that block
        ssi = self.feature_forward(im, spatial_block_width=self.spatial_box_width).reduce(force_numpy=True)

        lc_interpretable_napari('block_scaled_sum_intensity', ssi, viewer, im.ndim, ['ncells'])


# ---------------------------Convert Binary Mask to Instance Mask------------------------------


class DirectBSToOS(BlockToBlockProcess):
    def __init__(self):
        super().__init__(DatType.BS, DatType.OS, is_label=True)

    def np_forward(self, bs: np.ndarray[np.uint8]) -> np.ndarray[np.int32]:
        lbl_im, nlbl = instance_label(bs)
        return lbl_im


class Watershed3SizesBSToOS(BlockToBlockProcess):
    def __init__(self,
                 size_thres=60.,
                 dist_thres=1.,
                 rst=None,
                 size_thres2=100.,
                 dist_thres2=1.5,
                 rst2=60.):
        super().__init__(DatType.BS, DatType.OS, is_label=True)
        self.size_thres = size_thres
        self.dist_thres = dist_thres
        self.rst = rst
        self.size_thres2 = size_thres2
        self.dist_thres2 = dist_thres2
        self.rst2 = rst2

    def np_forward(self, bs: np.ndarray[np.uint8]) -> np.ndarray[np.int32]:
        lbl_im = algorithms.round_object_detection_3sizes(bs,
                                                          size_thres=self.size_thres,
                                                          dist_thres=self.dist_thres,
                                                          rst=self.rst,
                                                          size_thres2=self.size_thres2,
                                                          dist_thres2=self.dist_thres2,
                                                          rst2=self.rst2,
                                                          remap_indices=True)
        return lbl_im
    # TODO: better visualization of this stage


# ---------------------------Convert Binary Mask to Instance Mask------------------------------


class BinaryAndCentroidListToInstance(SegProcess):
    """Defines a SegProcess

    This class' instance forward() takes two inputs: The binary mask segmenting objects, and centroids
    detected by the blobdog algorithm. Then it splits the pixels in the binary mask into instance
    segmentation based on the centroids found by blobdog. The result is a more finely segmented mask
    where objects closer together are more likely to be correctly segmented as two
    """

    def __init__(self):
        super().__init__(DatType.OTHER, DatType.OS)

    def bacl_forward(self,
                     bs: np.ndarray[np.uint8],
                     lc: np.ndarray[np.float32],
                     block_info=None) -> np.ndarray[np.int32]:
        """For a numpy block and list of centroids in block, return segmentation based on centroids"""

        assert isinstance(bs, np.ndarray) and isinstance(lc, np.ndarray), \
            f'Error: inputs must be numpy for the forward() of this class, got bs={type(bs)} and lc={type(lc)}'

        # first sort each centroid into contour they belong to
        input_slices = block_info[0]['array-location']
        lc = lc.astype(np.int64) - np.array(tuple(s.start for s in input_slices), dtype=np.int64)[None, :]
        lbl_im, max_lbl = instance_label(bs)

        lbl_im: np.ndarray[np.int64]
        max_lbl: int

        # Below, index 0 is background - centroids fall within this are discarded
        contour_centroids = [[] for _ in range(max_lbl + 1)]

        for centroid in lc:
            c_ord = int(lbl_im[tuple(centroid)])
            contour_centroids[c_ord].append(centroid)

        def map_fn(
                centroids: list[np.ndarray[np.int64]],
                indices: list[int],
                X: tuple[np.ndarray]
        ) -> np.ndarray[np.int32]:
            N = len(centroids)
            assert N >= 2
            assert N == len(indices)
            arr_shape = X[0].shape
            X = np.array(X)
            indices = np.array(indices, dtype=np.int32)

            idxD = np.zeros(arr_shape, dtype=np.int32)
            minD = np.ones(arr_shape, dtype=np.float32) * 1e10
            for i in range(N):
                centroid = centroids[i]
                D = X - np.expand_dims(centroid, list(range(1, X.ndim)))
                D = np.linalg.norm(D.astype(np.float32), axis=0)  # euclidean distance
                new_mask = D < minD
                idxD = new_mask * indices[i] + ~new_mask * idxD
                minD = new_mask * D + ~new_mask * minD
            return idxD

        # now we compute the contours, and brute-force calculate what centroid each pixel is closest to
        object_slices = list(find_objects(lbl_im))
        new_lbl = max_lbl + 1

        for i in range(1, max_lbl + 1):
            slices = object_slices[i - 1]
            if slices is None:
                continue

            # if there are 0 or 1 centroid in the contour, we do nothing
            centroids = contour_centroids[i]  # centroids fall within the current contour
            ncentroid = len(centroids)
            if ncentroid <= 1:
                continue

            # otherwise, divide the contour and map pixels to each
            indices = [i] + [lbl for lbl in range(new_lbl, new_lbl + ncentroid - 1)]
            new_lbl += ncentroid - 1
            mask = lbl_im[slices] == i
            stpt = np.array(tuple(s.start for s in slices), dtype=np.int64)
            centroids = [centroid - stpt for centroid in centroids]
            divided = algorithms.coord_map(mask.shape,
                                           lambda *X: map_fn(centroids, indices, X))

            lbl_im[slices] = lbl_im[slices] * ~mask + divided * mask
        return lbl_im

    def forward(self, bs: np.ndarray[np.uint8] | da.Array, lc: NDBlock[np.float32]) \
            -> np.ndarray[np.int32] | da.Array:
        bs = NDBlock(bs)
        is_numpy = bs.is_numpy() or lc.is_numpy()
        if is_numpy:
            # if one is numpy but not both numpy, force both of them to be numpy
            if not lc.is_numpy():
                lc = lc.reduce(force_numpy=True)
            elif not bs.is_numpy():
                bs = bs.as_numpy()

        ndblock = NDBlock.map_ndblocks([bs, lc], self.bacl_forward,
                                       out_dtype=np.int32, use_input_index_as_arrloc=0)
        if is_numpy:
            return ndblock.as_numpy()
        else:
            return ndblock.as_dask_array()

    def interpretable_napari(self,
                             viewer: napari.Viewer,
                             bs: np.ndarray[np.uint8] | da.Array,
                             lc: NDBlock[np.float32]):
        # first sort each centroid into contour they belong to
        if isinstance(bs, np.ndarray):
            lbl_im = instance_label(bs)[0]
        else:
            lbl_im = bs.map_blocks(
                lambda block: instance_label(block)[0].astype(np.int32), dtype=np.int32
            )
            lbl_im = self.cache_im(lbl_im)
        viewer.add_labels(lbl_im, name='Inst seg before centroid split')

        after_split = self.cache_im(self.forward(bs, lc))
        after_split_layer = viewer.add_labels(after_split,
                                              name='Inst seg by pixel distance to centroid')


# ---------------------------Ordinal Segmentation to List of Centroids-------------------------


class DirectOSToLC(SegProcess):
    """Convert a 0-N contour mask to a list of N centroids, one for each contour

    The converted list of centroids is in the same order as the original contour order (The contour labeled
    1 will come first and before contour labeled 2, 3 and so on)
    """

    def __init__(self, reduce=False):
        super().__init__(DatType.OS, DatType.LC)
        self.reduce = reduce

    def np_features(self, block: np.ndarray[np.int32], block_info=None) \
            -> np.ndarray[np.float32]:
        if block_info is not None:
            slices = block_info[0]['array-location']
            contours_np3d = algorithms.npindices_from_os(block)
            lc = [contour.astype(np.float32).mean(axis=0) for contour in contours_np3d]
            lc = np.array(lc, dtype=np.float32)
            if slices is not None:
                start_pos = np.array([slices[i].start for i in range(len(slices))], dtype=np.float32)
                lc[:, :block.ndim] += start_pos[None, :]
            return lc
        else:
            return block

    def feature_forward(self, im: np.ndarray[np.int32] | da.Array) -> NDBlock[np.float32]:
        return NDBlock.map_ndblocks([NDBlock(im)], self.np_features, out_dtype=np.float32)

    def forward(self, im: np.ndarray[np.int32] | da.Array) -> NDBlock[np.float32]:
        ndblock = self.feature_forward(im)
        if self.reduce:
            ndblock = ndblock.reduce(force_numpy=False)
        return ndblock

    def interpretable_napari(self, viewer: napari.Viewer, im: np.ndarray[np.int32] | da.Array):
        features = self.feature_forward(im).reduce(force_numpy=True)
        lc_interpretable_napari('os_to_lc_centroids', features, viewer, im.ndim, [])


# -----------------------Convert List of Centroids to Cell Count Estimate----------------------


class CountLCEdgePenalized(SegProcess):
    """From a list of cell centroid locations, calculate a cell count estimate
    
    You need to provide an image_shape parameter due to the fact that lc does not contain
    information about input image shape

    Each centroid is simply treated as 1 cell when they are sufficiently far from the edge,
    but as they get closer to the edge the divisor becomes >1. and their estimate decreases
    towards 0, since cells near the edge may be double-counted (or triple or more counted
    if at a corner etc.)
    """

    def __init__(self,
                 im_shape: np.ndarray | tuple[int],
                 border_params: tuple[float, float, float] = (3., -.5, 2.),
                 reduce: bool = False):
        """Initialize a CountLCEdgePenalized object

        Args:
            im_shape: Shape of the blocks where rows in list of centroids locate in
            border_params: Specify how the cells on the border gets discounted. Formula is:
                intercept, dist_coeff, div_max = self.border_params
                mults = 1 / np.clip(intercept - border_dists * dist_coeff, 1., div_max)
                cc_list = np.prod(mults, axis=1)
            reduce: If True, reduce the results into a Numpy 2d array calling forward()
        """
        super().__init__(DatType.LC, DatType.CC)
        self.im_shape = np.array(im_shape, dtype=np.float32)
        self.border_params = border_params
        self.reduce = reduce

        intercept, dist_coeff, div_max = border_params
        assert intercept >= 1., f'intercept has to be >= 1. as divisor must be >= 1! (intercept={intercept})'
        assert dist_coeff <= 0., (f'The dist_coeff needs to be non-positive so divisor decreases as cell is further '
                                  f'from the edge')
        assert div_max >= 1., f'The divisor is >= 1, but got div_max < 1! (div_max={div_max})'

    def cc_list(self, lc: np.ndarray[np.float32]) -> np.ndarray[np.float32]:
        """Returns a cell count estimate for each contour in the list of centroids

        Args:
            lc: The list of centroids to be given cell estimates for

        Returns:
            A 1-d list, each element is a scalar cell count for the corresponding contour centroid in lc
        """
        midpoint = (self.im_shape * .5)[None, :]

        # compute border distances in each axis direction
        border_dists = np.abs((lc + midpoint) % self.im_shape - (midpoint - .5))

        intercept, dist_coeff, div_max = self.border_params
        mults = 1 / np.clip(intercept + border_dists * dist_coeff, 1., div_max)
        cc_list = np.prod(mults, axis=1)
        return cc_list

    def np_features(self, lc: np.ndarray[np.float32], block_info=None) -> np.ndarray[np.float32]:
        """Calculate cell counts, then concat centroid locations to the left of cell counts"""
        cc_list = self.cc_list(lc)
        features = np.concatenate((lc, cc_list[:, None]), axis=1)
        return features

    def feature_forward(self, lc: NDBlock[np.float32]) -> NDBlock[np.float32]:
        return NDBlock.map_ndblocks([lc], self.np_features, out_dtype=np.float32)

    def forward(self, lc: NDBlock[np.float32]) -> NDBlock[np.float32]:
        ndblock = self.feature_forward(lc).select_columns([-1])
        if self.reduce:
            ndblock = ndblock.reduce(force_numpy=False)
        return ndblock.sum(keepdims=True)

    def interpretable_napari(self, viewer: napari.Viewer, lc: NDBlock[np.float32]):
        features = self.feature_forward(lc).reduce(force_numpy=True)
        lc_interpretable_napari('edge_penalized_centroids', features, viewer,
                                len(self.im_shape), ['ncells'])


# --------------------------Convert Ordinal Mask to Cell Count Estimate------------------------


class CountOSBySize(SegProcess):
    """Counting ordinal segmentation contours

    Several features:
    1. A size threshold, below which each contour is counted as a single cell (or part of a single cell,
        in the case it is neighbor to boundary of the image)
    2. Above size threshold, the contour is seen as a cluster of cells an estimate of cell count is given
        based on the volume of the contour
    3. For cells on the boundary location, their estimated ncell is penalized according to the distance
        between the cell centroid and the boundary of the image; if the voxels of the cell do not touch
        edge, this penalty does not apply
    4. A min_size threshold, below (<=) which the contour is simply discarded because it's likely just
        an artifact
    """

    def __init__(self,
                 size_threshold: int | float = 25.,
                 volume_weight: float = 6e-3,
                 border_params: tuple[float, float, float] = (3., -.5, 2.),
                 min_size: int | float = 0,
                 reduce: bool = False):
        super().__init__(DatType.OS, DatType.CC)
        self.size_threshold = size_threshold
        self.volume_weight = volume_weight
        self.border_params = border_params
        self.min_size = min_size
        self.reduce = reduce

    def cc_list(self, os: np.ndarray[np.int32]) -> np.ndarray[np.float32]:
        contours_np3d = algorithms.npindices_from_os(os)
        ncells = {}
        dc = []
        dc_idx_to_centroid_idx = {}
        idx_max = np.array(tuple(d - 1 for d in os.shape), dtype=np.int64)
        for i in range(len(contours_np3d)):
            contour = contours_np3d[i]
            nvoxel = contour.shape[0]
            if nvoxel <= self.min_size:
                ncells[i] = 0.
            else:
                ncells[i] = 0.

                # if no voxel touch the boundary, we do not want to apply the edge penalty
                on_edge = (contour == 0).astype(np.uint8) + (contour == idx_max[None, :]).astype(np.uint8)
                on_edge = on_edge.sum().item() > 0
                if on_edge:
                    dc_idx_to_centroid_idx[len(dc)] = i
                    dc.append(contour.astype(np.float32).mean(axis=0))
                else:
                    ncells[i] = 1

                if nvoxel > self.size_threshold:
                    ncells[i] += (nvoxel - self.size_threshold) * self.volume_weight
        ps = CountLCEdgePenalized(os.shape, self.border_params)
        dc_centroids = np.array(dc, dtype=np.float32)
        dc_ncells = ps.cc_list(dc_centroids)
        for dc_idx in dc_idx_to_centroid_idx:
            i = dc_idx_to_centroid_idx[dc_idx]
            ncells[i] += dc_ncells[dc_idx]
        ncells = np.array([ncells[i] for i in range(len(ncells))], dtype=np.float32)
        return ncells

    def np_features(self,
                    os: np.ndarray[np.int32],
                    block_info=None) -> np.ndarray[np.float32]:
        if block_info is None:
            return np.zeros(tuple(), dtype=np.float32)

        # slices = block_info[0]['array-location']
        ps = DirectOSToLC(reduce=True)
        lc = ps.np_features(os, block_info)
        cc_list = self.cc_list(os)
        features = np.concatenate((lc, cc_list[:, None]), axis=1)
        return features

    def feature_forward(self, im: np.ndarray | da.Array) -> NDBlock[np.float32]:
        return NDBlock.map_ndblocks([NDBlock(im)], self.np_features, out_dtype=np.float32)

    def forward(self, im: np.ndarray | da.Array) \
            -> np.ndarray:
        ndblock = self.feature_forward(im).select_columns([-1])
        if self.reduce:
            ndblock = ndblock.reduce(force_numpy=False)
        return ndblock.sum(keepdims=True)

    def interpretable_napari(self, viewer: napari.Viewer, im: np.ndarray[np.float32]):
        features = self.feature_forward(im).reduce(force_numpy=True)
        lc_interpretable_napari('edge_penalized_centroids', features, viewer, im.ndim, ['ncells'])



"""
This file provides visualization utilities for ome-zarr file, similar to napari-ome-zarr,
but includes features like displaying zip ome zarr files
"""

import napari
import zarr
import dask.array as da
from cvpl_tools.ome_zarr.io import load_zarr_group_from_path


# -------------Part 1: convenience functions, for adding ome zarr images using paths--------------


def add_ome_zarr_group_from_path(viewer: napari.Viewer, path: str, use_zip: bool | None = None, kwargs=None):
    """Add an ome zarr group to napari viewer from given group path.

    This is a combination of load_zarr_group_from_path() and add_ome_zarr_group() functions.
    """
    zarr_group = load_zarr_group_from_path(path, 'r', use_zip)
    add_ome_zarr_group(viewer, zarr_group, kwargs)


def add_ome_zarr_array_from_path(viewer: napari.Viewer, path: str, use_zip: bool | None = None, kwargs=None):
    """Add an ome zarr array to napari viewer from given array path.

    This is a combination of load_zarr_array_from_path() and add_ome_zarr_group() functions.
    """
    if kwargs is None:
        kwargs = {}
    zarr_group = load_zarr_group_from_path(path, 'r', use_zip)
    add_ome_zarr_array(viewer, zarr_group, **kwargs)


# ------------------------Part 2:adding ome zarr files using zarr group---------------------------


def add_ome_zarr_group(viewer: napari.Viewer, zarr_group: zarr.hierarchy.Group, kwargs=None):
    """Add an ome zarr image (if exists) along with its labels (if exist) to viewer.

    Args
        viewer: Napari viewer object to attach image to
        zarr_group: The zarr group that contains the ome zarr file
        kwargs: dictionary, keyword arguments to be passed to viewer.add_image for root image
    """
    if kwargs is None:
        kwargs = {}
    if '0' in zarr_group:
        add_ome_zarr_array(viewer, zarr_group, **kwargs)
    if 'labels' in zarr_group:
        lbls_group = zarr_group['labels']
        for group_key in lbls_group.group_ordered_keys():
            lbl_group = lbls_group[group_key]
            add_ome_zarr_array(viewer, lbl_group, name=group_key)


def add_ome_zarr_array(viewer: napari.Viewer, zarr_group: zarr.hierarchy.Group, start_level: int = 0, **kwargs):
    """Add a multiscale ome zarr image or label to viewer.

    Args:
        viewer (napari.Viewer): Napari viewer object to attach image to.
        zarr_group (zarr.hierarchy.Group): The zarr group that contains the ome zarr file.
        start_level (int): The lowest level (highest resolution) to be added, default to 0
        ``**kwargs``: Keyword arguments to be passed to viewer.add_image for root image.
    """
    multiscale = []
    while True:
        i = len(multiscale) + start_level
        i_str = str(i)
        if i_str in zarr_group:  # by ome zarr standard, image pyramid starts from 0 to NLEVEL - 1
            multiscale.append(da.from_zarr(zarr_group[i_str]))
        else:
            break
    viewer.add_image(multiscale, multiscale=True, **kwargs)



"""
This file defines some helper functions and algorithms for parallel dask image processing
"""
from __future__ import annotations

import abc
import copy
import enum
from typing import Callable, Iterator, Iterable, Sequence, Any, Generic, TypeVar

import numpy as np
import dask.array as da
import dask
import functools
import operator


def _init_ndlist(idx: tuple, shape: tuple, init_element: Callable[[tuple], Any]) -> list:
    if len(shape) > 1:
        return [_init_ndlist(idx + (i,), shape[1:], init_element) for i in range(shape[0])]

    return [init_element(idx + (i,)) for i in range(shape[0])]


def _dask_ndconcat(ndlists: Sequence, ax_orders: Sequence[int] | None = None):
    if len(ax_orders) > 1:
        sub_ax_orders = ax_orders[1:]
        ndlists = [_dask_ndconcat(ndlists[i], sub_ax_orders) for i in range(len(ndlists))]
    # TODO: This does not parallelize well, how can this be delayed?
    return da.concatenate(da.compute(*ndlists), axis=ax_orders[0])


def dask_ndconcat(ndlists: Sequence, ndim: int | None = None, ax_orders: Sequence[int] | None = None):
    """Concatenate a ndlists of dask delayed into a single dask image

    This function will correctly handle both when blocks are numpy ndarray or when they are dask arrays
    """
    assert not (ndim is None and ax_orders is None), ('Expected to receive at least one of ndim or ax_order as input, '
                                                      f'got ndim={ndim} and ax_orders={ax_orders}')
    if ndim is not None and ax_orders is not None:
        assert len(ax_orders) == ndim, f'Conflicting inputs, got ndim={ndim} and ax_orders={ax_orders}'

    if ax_orders is None:
        ax_orders = tuple(range(ndim))
    return _dask_ndconcat(ndlists, ax_orders)


class ReprFormat(enum.Enum):
    NUMPY_ARRAY = 0
    DASK_ARRAY = 1
    DICT_BLOCK_INDEX_SLICES = 2  # block can be either numpy or dask


ElementType = TypeVar('ElementType')


class NDBlock(Generic[ElementType], abc.ABC):
    """This class represents a nd arrangement of arbitrarily sized blocks

    Currently, we assume block_index is a list always ordered in (0, ..., 0), (0, ..., 1)... (N, ..., M-1) Increasing
    from the side of the tail first
    """

    def __init__(self, arr: np.ndarray | da.Array | NDBlock | None):
        self.arr = None
        self.properties = None

        if arr is None:
            return  # don't fully initialize; used in case properties can't be inferred from arr

        if isinstance(arr, np.ndarray):
            self.arr = arr
            self._set_properties_by_numpy_array(arr)
        elif isinstance(arr, da.Array):
            self.arr = arr
            self._set_properties_by_dask_array(arr)
        else:
            assert isinstance(arr, NDBlock)
            self.arr = arr.arr
            self.properties = NDBlock._copy_properties(arr.properties)

    @staticmethod
    def from_ndlists(ndlists: Sequence):
        return NDBlock(dask_ndconcat(ndlists))

    def is_numpy(self) -> bool:
        return self.properties['is_numpy']

    def get_repr_format(self) -> ReprFormat:
        return self.properties['repr_format']

    def get_ndim(self) -> int:
        return self.properties['ndim']

    def get_numblocks(self):
        return self.properties['numblocks']

    def get_block_indices(self) -> list:
        return self.properties['block_indices']

    def get_dtype(self):
        return self.properties['dtype']

    def as_numpy(self) -> da.Array:
        if self.properties['repr_format'] == ReprFormat.NUMPY_ARRAY:
            return self.arr
        else:
            other = NDBlock(self)
            other.to_numpy_array()
            return other.arr

    def as_dask_array(self) -> da.Array:
        if self.properties['repr_format'] == ReprFormat.DASK_ARRAY:
            return self.arr
        else:
            other = NDBlock(self)
            other.to_dask_array()
            return other.arr

    @staticmethod
    def _copy_properties(properties: dict):
        return dict(
            repr_format=properties['repr_format'],
            ndim=properties['ndim'],
            numblocks=properties['numblocks'],
            block_indices=properties['block_indices'],
            slices_list=[item for item in properties['slices_list']],
            is_numpy=properties['is_numpy'],
            dtype=properties['dtype']
        )

    def _set_properties_by_numpy_array(self, arr: np.ndarray):
        ndim: int = arr.ndim
        self.properties = dict(
            repr_format=ReprFormat.NUMPY_ARRAY,
            ndim=ndim,
            numblocks=(1,) * ndim,
            block_indices=(0,) * ndim,
            slices_list=list((0, s) for s in arr.shape),
            is_numpy=True,
            dtype=arr.dtype
        )

    def _set_properties_by_dask_array(self, arr: da.Array):
        ndim: int = arr.ndim
        numblocks = arr.numblocks
        block_indices = list(np.ndindex(*arr.numblocks))
        slices_list: list = da.core.slices_from_chunks(arr.chunks)

        self.properties = dict(
            repr_format=ReprFormat.DASK_ARRAY,
            ndim=ndim,
            numblocks=numblocks,
            block_indices=block_indices,
            slices_list=slices_list,
            is_numpy=False,
            dtype=arr.dtype
        )

    # ReprFormat conversion functions

    def to_numpy_array(self):
        """Convert representation format to numpy array"""
        if self.properties['repr_format'] == ReprFormat.DICT_BLOCK_INDEX_SLICES:
            self.to_dask_array()

        rformat = self.properties['repr_format']
        if rformat == ReprFormat.NUMPY_ARRAY:
            return

        assert rformat == ReprFormat.DASK_ARRAY
        self.arr = self.arr.compute()
        self._set_properties_by_numpy_array(self.arr)  # here some blocks may be merged to one

        self.properties['repr_format'] = ReprFormat.NUMPY_ARRAY
        self.properties['is_numpy'] = True

    def to_dask_array(self):
        """Convert representation format to dask array"""
        rformat = self.properties['repr_format']
        if rformat == ReprFormat.DASK_ARRAY:
            return

        if rformat == ReprFormat.NUMPY_ARRAY:
            self.arr = da.from_array(self.arr)
        else:
            assert rformat == ReprFormat.DICT_BLOCK_INDEX_SLICES
            numblocks = self.get_numblocks()
            dtype = self.get_dtype()

            # def extract_block(idx: tuple[int]):
            #     block = self.arr[idx][0]
            #     if not isinstance(block, np.ndarray):
            #         block = da.from_delayed(block, shape=block_shape, dtype=dtype)
            #     return block

            # reference: https://github.com/dask/dask-image/blob/adcb217de766dd6fef99895ed1a33bf78a97d14b/dask_image/ndmeasure/__init__.py#L299
            ndlists = np.empty(numblocks, dtype=object)
            for block_index in self.arr.keys():
                block, slices = self.arr[block_index]
                block_shape = tuple(s.stop - s.start for s in slices)
                if not isinstance(block, np.ndarray):
                    block = da.from_delayed(block, shape=block_shape, dtype=dtype)
                ndlists[block_index] = block
            ndlists = ndlists.tolist()
            self.arr = da.block(ndlists)
        self.properties['repr_format'] = ReprFormat.DASK_ARRAY
        self.properties['is_numpy'] = False

    def to_dict_block_index_slices(self):
        rformat = self.properties['repr_format']
        if rformat == ReprFormat.DICT_BLOCK_INDEX_SLICES:
            return

        slices_list = self.properties['slices_list']
        block_indices = self.properties['block_indices']
        arr = self.arr
        if rformat == ReprFormat.NUMPY_ARRAY:
            blocks = [arr]
        else:
            assert isinstance(arr, da.Array)
            blocks = list(map(functools.partial(operator.getitem, arr), slices_list))
        nblocks = len(blocks)
        assert len(slices_list) == nblocks and len(block_indices) == nblocks, \
            (f'Expected equal length of block, indices and slices, got len(blocks)={nblocks} '
             f'len(block_indices)={len(block_indices)} len(slices_list)={len(slices_list)}')

        self.arr = {
            block_indices[i]: (blocks[i], slices_list[i]) for i in range(len(blocks))
        }

        self.properties['repr_format'] = ReprFormat.DICT_BLOCK_INDEX_SLICES

    def reduce(self, force_numpy: bool = False) -> np.ndarray | da.Array:
        """concatenate all blocks on the first axis"""
        if self.properties['repr_format'] == ReprFormat.NUMPY_ARRAY:
            return np.copy(self.arr)
        else:
            other = NDBlock(self)
            other.to_dict_block_index_slices()

            shape = None
            for block_index, (block, slices) in other.arr.items():
                if not isinstance(block, np.ndarray):
                    block = block.compute()
                shape = (np.nan,) + block.shape[1:]
                break
            blocks = [da.from_delayed(block,
                                      shape=shape,
                                      dtype=self.get_dtype())
                      for block_index, (block, slices) in other.arr.items()]
            assert len(blocks) > 0, 'Need at least one row for NDBlock to be reduced'
            if other.is_numpy():
                return np.concatenate(blocks, axis=0)
            else:
                reduced = da.concatenate(blocks, axis=0)
                if force_numpy:
                    return reduced.compute()
                else:
                    return reduced

    def sum(self, axis: Sequence | None = None, keepdims: bool = False):
        """sum over axes for each block"""
        new_ndblock = NDBlock(self)
        if self.properties['repr_format'] == ReprFormat.DICT_BLOCK_INDEX_SLICES:
            for block_index, (block, slices) in self.arr.items():
                new_ndblock.arr[block_index] = (block.sum(axis=axis, keepdims=keepdims), slices)
            return new_ndblock
        else:
            new_ndblock.arr = self.arr.sum(axis=axis)

    @staticmethod
    def map_ndblocks(
        inputs: Sequence[NDBlock],
        fn: Callable,
        out_dtype: np.dtype,
        use_input_index_as_arrloc: int = 0
    ) -> NDBlock:
        """similar to da.map_blocks but works with NDBlock

        Args:
            inputs: A list of inputs to be mapped, either all dask or all numpy; All inputs must have the same number
                of blocks, block indices, and over the same slices as well if inputs are dask images
            fn: fn(*block, block_info) maps input blocks to an output block
            use_input_index_as_arrloc: output slices_list will be the same as this input (ignore this for variable
                sized blocks)

        Returns:
            Result mapped, of format ReprFormat.DICT_BLOCK_INDEX_SLICES
        """
        inputs = list(inputs)
        assert len(inputs) > 0, 'Must have at least one input!'

        is_numpy = inputs[0].is_numpy()
        for i in range(1, len(inputs)):
            assert is_numpy == inputs[i].is_numpy(), ('All inputs must either be all dask or all Numpy, expected '
                                                      f'is_numpy={is_numpy} but found wrong typed input at '
                                                      f'location {i}')

        if is_numpy:
            block_info = []
            for i in range(len(inputs)):
                ndblock = inputs[i]
                arr = ndblock.as_numpy()
                block_info.append({
                    'chunk-location': (0,) * arr.ndim,
                    'array-location': list((0, s) for s in arr.shape)
                })
            return NDBlock(fn(*inputs, block_info))

        # we can assume the inputs are all dask, now turn them all into block iterator
        block_iterators = []
        for ndblock in inputs:
            if ndblock.get_repr_format() != ReprFormat.DICT_BLOCK_INDEX_SLICES:
                new_block = NDBlock(ndblock)
                new_block.to_dict_block_index_slices()
                block_iterators.append(new_block)
            else:
                block_iterators.append(ndblock)

        @dask.delayed
        def delayed_fn(*blocks, block_info):
            return fn(*blocks, block_info)

        result = {}
        for block_index in block_iterators[use_input_index_as_arrloc].arr.keys():
            blocks = []
            block_info = []
            for i in range(len(block_iterators)):
                block, slices = block_iterators[i].arr[block_index]
                blocks.append(block)
                block_info.append({
                    'chunk-location': block_index,
                    'array-location': slices,
                })
            result[block_index] = (delayed_fn(*blocks, block_info=block_info),
                                   block_info[use_input_index_as_arrloc]['array-location'])

        ndim = block_iterators[0].get_ndim()
        numblocks = block_iterators[0].get_numblocks()
        result_ndblock = NDBlock(None)
        result_ndblock.arr = result
        result_ndblock.properties = dict(
            repr_format=ReprFormat.DICT_BLOCK_INDEX_SLICES,
            ndim=ndim,
            numblocks=numblocks,
            block_indices=block_iterators[use_input_index_as_arrloc].arr.keys(),
            slices_list=block_iterators[use_input_index_as_arrloc].properties['slices_list'],
            is_numpy=False,
            dtype=out_dtype
        )
        return result_ndblock

    def select_columns(self, cols: slice | Sequence[int] | int) -> NDBlock:
        """Performs columns selection on a 2d array"""
        if self.properties['repr_format'] == ReprFormat.DICT_BLOCK_INDEX_SLICES:
            results = {
                block_index: (block[:, cols], slices) for block_index, (block, slices) in self.arr.items()
            }
            ndblock = NDBlock(self)
            ndblock.arr = results
        else:
            ndblock = NDBlock(self)
            ndblock.arr = ndblock.arr[:, cols]
        return ndblock




"""
fs.py
This is a file that provides methods to interface with image files.

Stored Numpy arrays in computer vision can be represented in .npy, .tif, .png, .jpg etc.
And when reading and writing files, the axis order and pixel format (grey vs. RGB vs RGBA)
often complicate the code.

This file aims to provide a unified way to manipulate, read and write np arrays based on a
file path pattern e.g. "./dataset1/dog_*.png" and a few settings specifying the reading
axis order and underlying data arrangement formats
"""


import os
import shutil
import glob
import numpy as np
import nibabel as nib
from PIL import Image
import scipy
from dataclasses import dataclass, field
from cvpl_tools.array_key_dict import ArrayKeyDict
from nibabel import load as nib_load


def ensure_dir_exists(dir_path, remove_if_already_exists):
    """
    If a directory does not exist, make a new directory with the name.
    This assumes the parent directory must exists; otherwise a path not
    found error will be thrown.
    Args:
        dir_path: The path of folder
        remove_if_already_exists: if True and the folder already exists, then remove it and make a new one.
    """
    if os.path.exists(dir_path):
        if remove_if_already_exists:
            shutil.rmtree(dir_path)
            os.mkdir(dir_path)
    else:
        os.mkdir(dir_path)


class ImFileType:
    """
    Represents different file types that can be read and written by fs.py
    """
    FTYPE_UNKNOWN = -1
    FTYPE_JPX = 0
    FTYPE_JPG = 1
    FTYPE_PNG = 2
    FTYPE_NIB = 3
    FTYPE_TIF = 4
    FTYPE_NPY = 5
    FTYPE_NPZ = 6  # probabilities output by nnunet
    ext_to_ftype = {
        '.jpx': FTYPE_JPX,
        '.jpg': FTYPE_JPG,
        '.png': FTYPE_PNG,
        '.nii.gz': FTYPE_NIB,
        '.tif': FTYPE_TIF,
        '.npy': FTYPE_NPY,
        '.npz': FTYPE_NPZ,
    }
    ftype_to_ext = list(ext_to_ftype.keys())
    all_ftypes = list(ext_to_ftype.values())  # except unknown

    FORMAT_UINT8 = 0
    FORMAT_UINT16 = 1
    FORMAT_INT32 = 2
    FORMAT_FLOAT32 = 3
    format_to_np_dtype = [np.uint8, np.uint16, np.int32, np.float32]

    @staticmethod
    def ftype_from_im_path(im_path: str) -> int:
        """
        Infer the ftype from image path;
        The path can contain wildcards like "*"
        When the ftype can't be inferred, FTYPE_UNKNOWN will be returned
        """
        prefix, ext = os.path.splitext(im_path)
        ext = ext.lower()
        if ext == '.gz':
            prefix_prefix, prefix_ext = os.path.splitext(prefix)
            if prefix_ext.lower() == '.nii':
                return ImFileType.FTYPE_NIB
            else:
                return ImFileType.FTYPE_UNKNOWN
        elif ext in ImFileType.ext_to_ftype:
            return ImFileType.ext_to_ftype[ext]
        else:
            return ImFileType.FTYPE_UNKNOWN

    @staticmethod
    def get_imid_and_zstack(im_path: str, is_stack=False, keep_dirpath=False) -> tuple[str, str]:
        """
        Get the path stem as the unique image id from the path of the file;
        Second return value is the zstack number as a string
        file extensions will be stripped;
        If the file is part of a stacked image, then the z-stack numbers will be stripped from stem as will
        """
        ftype = ImFileType.ftype_from_im_path(im_path)
        if ftype == ImFileType.FTYPE_UNKNOWN:
            raise ValueError(f'Attempt to get the imid of an image file from path "{im_path}" with unknown file type!')
        elif ftype == ImFileType.FTYPE_NIB:
            imid = im_path[:-len('.nii.gz')]
        else:
            imid = os.path.splitext(im_path)[0]

        if not keep_dirpath:
            imid = os.path.split(imid)[1]

        if is_stack:
            i = -1
            while len(imid) >= -i and imid[i].isdigit():
                i -= 1
            return imid[:i + 1], imid[i + 1:]
        else:
            return imid, ''

    @staticmethod
    def get_imid(im_path: str, is_stack=False, keep_dirpath=False) -> str:
        return ImFileType.get_imid_and_zstack(im_path, is_stack, keep_dirpath)[0]

    @staticmethod
    def load_im_as_np(im_format, path, true_im_ndim: int, stack_axis=None,
                      concat_instead=False, allow_pickle=False) -> tuple:
        """
        Load individual file as a np array, or a stack of images as a np array if a list of path is given
        Returns a tuple (im, im_ftype, im_meta) where im_meta may be needed in order to save the im back to a file
        for certain file types
        """
        if isinstance(path, str):
            ftype = ImFileType.ftype_from_im_path(path)
            assert stack_axis is None, 'ERROR: Passing stack_axis in when loading single file'
            assert ftype != ImFileType.FTYPE_UNKNOWN, 'ERROR: Attempt to load an unknown image type!'
            if ftype == ImFileType.FTYPE_NPZ:
                assert im_format == ImFileType.FORMAT_FLOAT32
                im = np.load(path, allow_pickle=allow_pickle)['probabilities']
                im = np.transpose(im, (0, 3, 2, 1))
                im_meta = [ftype, im.shape, None, path, None]
            elif ftype == ImFileType.FTYPE_NPY:
                im = np.load(path, allow_pickle=allow_pickle)
                if allow_pickle:
                    im = im.item()['masks']
                im_meta = [ftype, im.shape, None, path, None]
            elif ftype == ImFileType.FTYPE_NIB:
                nii_im = nib_load(path)
                im = np.array(nii_im.get_fdata(), dtype=ImFileType.format_to_np_dtype[im_format])
                im_meta = [ftype, im.shape, None, path, nii_im.affine]
            else:
                im = np.array(Image.open(path), dtype=ImFileType.format_to_np_dtype[im_format])
                im_meta = [ftype, im.shape, None, path, None]

            assert im.dtype == ImFileType.format_to_np_dtype[im_format], \
                f'ERROR: Expect file type {ImFileType.format_to_np_dtype[im_format]} but got an image of dtype {im.dtype} at {path}'
            if len(im.shape) > true_im_ndim:
                axes = range(true_im_ndim, len(im.shape))
                for i in axes:
                    assert im.shape[i] == 1, f'Extra axes to true_im_ndim must be of width 1. true_im_ndim={true_im_ndim}, im.shape={im.shape}'
                im = im.squeeze(axis=tuple(axes))
            assert len(im.shape) == true_im_ndim, (f'ERROR: Expected true dimension {true_im_ndim}, but found '
                                                  f'image of shape {im.shape}!')
            assert type(im_meta) is list, 'ERROR: im_meta is not properly initialized!'
            return im, im_meta
        else:
            assert isinstance(path, list)
            im_arr = []
            im_meta = None
            for filepath in path:
                if concat_instead:
                    true_slice_dim = true_im_ndim
                else:
                    true_slice_dim = true_im_ndim - 1
                tup = ImFileType.load_im_as_np(im_format, filepath, true_slice_dim)
                im_arr.append(tup[0])
                cur_im_meta = tup[1]
                if im_meta is None:
                    im_meta = list(cur_im_meta)  # shallow copy
                    ftype_meta = im_meta[-1]
                    im_meta[-1] = [ftype_meta]
                else:
                    assert im_meta[0] == cur_im_meta[0]  # same file type
                    assert im_meta[1] == cur_im_meta[1]  # each image has the same shape
                    im_meta[-1].append(cur_im_meta[-1])
            if stack_axis is None:
                raise ValueError(f"Please specify stack_axis!")
                # stack_axis = true_im_ndim - 1  # default stack on the last axis
            if concat_instead:  # concat on existing axis
                assert 0 <= stack_axis < true_im_ndim - 1, f'Attempt to concat on axis {stack_axis} with true dim={true_im_ndim}'
                im = np.concatenate(im_arr, axis=stack_axis)
            else:  # stack on new axis
                assert 0 <= stack_axis < true_im_ndim, f'Attempt to concat on axis {stack_axis} with true dim={true_im_ndim}'
                im = np.stack(im_arr, axis=stack_axis)

            im_meta[2] = stack_axis
            im_meta[3] = path

            return im, im_meta

    @staticmethod
    def save_np_as_im(target_path: str, im: np.ndarray, ftype=None, stack_axis=None,
                      concat_instead=False, ftype_meta=None, allow_pickle=False, im_format=None):
        """
        If ftype is None (default), it will be inferred from path
        """
        true_im_ndim = len(im.shape)
        if type(target_path) is str:
            assert not os.path.exists(target_path), f'{target_path} exists!'
            if ftype is None:
                ftype = ImFileType.ftype_from_im_path(target_path)
            if im_format is not None:
                assert im.dtype == ImFileType.format_to_np_dtype[im_format]
            assert ftype != ImFileType.FTYPE_UNKNOWN, 'ERROR: Attempt to save an unknown image type!'
            if ftype == ImFileType.FTYPE_NPZ:
                assert im.dtype == np.float32
                im = np.transpose(im, (0, 3, 2, 1))
                np.savez(target_path, probabilities=im, allow_pickle=allow_pickle)
            elif ftype == ImFileType.FTYPE_NPY:
                if allow_pickle:
                    im = {'masks': im}
                np.save(target_path, im, allow_pickle=allow_pickle)
            elif ftype == ImFileType.FTYPE_NIB:
                assert ftype_meta.shape == (4, 4), f'Affine should have shape 4, 4, but get {ftype_meta[-1].shape}'
                nii_im = nib.Nifti1Image(im, ftype_meta)
                nib.save(nii_im, target_path)
            else:
                Image.fromarray(im).save(target_path)
        else:
            if stack_axis is None:
                stack_axis = true_im_ndim - 1
            for i in range(im.shape[stack_axis]):
                if concat_instead:
                    im_slice = im.take(indices=range(i, i + 1), axis=stack_axis)
                else:
                    im_slice = im.take(indices=i, axis=stack_axis)
                target = target_path[i]
                if ftype_meta is None:
                    cur_ftype_meta = None
                else:
                    cur_ftype_meta = ftype_meta[i]
                # recursive call!
                ImFileType.save_np_as_im(
                    target_path=target,
                    im=im_slice,
                    ftype=ftype,
                    stack_axis=stack_axis,
                    concat_instead=False,
                    ftype_meta=cur_ftype_meta,
                    allow_pickle=allow_pickle)


@dataclass
class ImReadSetting:
    """
    im_format (int) - specifies the data storage layout e.g. ImFileType.FORMAT_UINT8
    stack_axis (int) - if not None, combine several files with different zstacks to the same imid; specify the axis to stack on
    concat_instead (bool) - if True, the stacking are done by np.concatenate() which preserves im dimensions; over the same axis
    true_im_ndim (int) - # of dimensions of a single image; includes number of channels (for 2d BGR image this would be 3)
    allow_pickle (bool) - set allow_pickle for np.load and np.save
    imid_keep_dirpath (bool) - keeping the directory path when extracting imid from paths; default to False
    """
    im_format: int = ImFileType.FORMAT_UINT8
    stack_axis: int | None = None
    concat_instead: bool = False
    true_im_ndim: int = 3
    allow_pickle: bool = False
    imid_keep_dirpath: bool = False


@dataclass
class ImWriteSetting:
    """
    im_format (int) - specifies the data storage layout e.g. ImFileType.FORMAT_UINT8
    stack_axis (int) - if not None, combine several files with different zstacks to the same imid; specify the axis to stack on
    concat_instead (bool) - if True, the stacking are done by np.concatenate() which preserves im dimensions; over the same axis
    allow_pickle (bool) - set allow_pickle for np.load and np.save
    ftype (int) - file type to write e.g. ImFileType.FTYPE_NPZ
    """
    im_format: int = ImFileType.FORMAT_UINT8
    stack_axis: int | None = None
    concat_instead: bool = False
    allow_pickle: bool = False
    ftype: int | None = None


class ImIO:
    @staticmethod
    def read_single_image(read_setting: ImReadSetting, path):
        im, im_meta = ImFileType.load_im_as_np(read_setting.im_format, path,
                                               true_im_ndim=read_setting.true_im_ndim,
                                               stack_axis=read_setting.stack_axis,
                                               concat_instead=read_setting.concat_instead,
                                               allow_pickle=read_setting.allow_pickle)
        return im, im_meta

    @staticmethod
    def write_single_image(write_setting: ImWriteSetting, path, im, im_meta):
        ImFileType.save_np_as_im(path, im=im, ftype=write_setting.ftype,
                                 stack_axis=write_setting.stack_axis,
                                 concat_instead=write_setting.concat_instead,
                                 ftype_meta=im_meta[-1],
                                 allow_pickle=write_setting.allow_pickle)

    @staticmethod
    def read_filenames(read_setting: ImReadSetting, pattern: str | None = None, path: list | None = None):
        """
        The keys (imids) of the ArrayKeyDict returned by this function will be sorted alphabetically
        Parameters
            path (list) - optional and passed in as a list of all paths if pattern is None
        Return
            imid_to_path (Dict) - maps from imids to either a filename or a list of filenames (if image is stacked);
                The keys of this dictionary is a sorted list of image imids, act as unique ids of images, sorted alphabetically
        """
        imid_to_path = ImIO._get_imid_to_path_mapping(read_setting, path, pattern)
        imid_to_path.reorder_keys(sorted(imid_to_path.ordered_keys()))
        return imid_to_path

    @staticmethod
    def read_image(read_setting: ImReadSetting, paths: list):
        """
        Parameter
            paths (list) - can be nested in case of stacked images
        Return
            an iterator whose elements are (im, im_meta), im is the np array of the image and im_meta contains much
            additional metadata returned by ImFileType.load_im_as_np()
        """
        assert type(paths) is not ArrayKeyDict, f'Attempt to pass an ArrayKeyDict to read_image()!'
        for path in paths:
            im, im_meta = ImIO.read_single_image(read_setting, path)
            yield im, im_meta

    @staticmethod
    def write_image(write_setting: ImWriteSetting, paths: list, im_iter: iter):
        """
        Parameter
            paths (list) - can be nested in case of stacked images
            im_iter (iter) - An iterator returning tuples (im, im_meta), the order of these images should be the same as
                the order of their paths in the paths variable
        """
        for path in paths:
            im, im_meta = next(im_iter)
            ImIO.write_single_image(write_setting, path, im, im_meta)

    @staticmethod
    def _get_imid_to_path_mapping(read_setting: ImReadSetting,
                                  paths: list[str] | None = None, pattern: str | None = None):
        if paths is None:
            paths = glob.glob(pattern)
        elif pattern is not None:
            print('WARNING: both paths and pattern are not None! Is this intended?')

        imid_to_path = ArrayKeyDict()
        for path in paths:
            ftype = ImFileType.ftype_from_im_path(path)
            if ftype == ImFileType.FTYPE_UNKNOWN:
                continue
            imid, zstack_str = ImFileType.get_imid_and_zstack(path, is_stack=read_setting.stack_axis is not None,
                                                              keep_dirpath=read_setting.imid_keep_dirpath)
            if zstack_str == '':
                imid_to_path[imid] = path
            else:
                tup = (zstack_str, path)
                if imid not in imid_to_path:
                    imid_to_path[imid] = [tup]
                else:
                    imid_to_path[imid].append(tup)
        for k in list(imid_to_path.ordered_keys()):
            v = imid_to_path[k]
            if type(v) is list:
                imid_to_path[k] = [tup[1] for tup in sorted(v)]
        return imid_to_path


def im_resize(target_size: tuple, im: np.ndarray, order: int = 0) -> np.ndarray:
    assert len(im.shape) == len(target_size), f'Image has shape {im.shape} but intend to resize ' \
                                              f'to incompatible shape {target_size}'
    zoom_factors = (target_size[i] / im.shape[i] for i in range(len(target_size)))
    resized_im = scipy.ndimage.zoom(im, zoom_factors, order=order)  # order=0 -> use nearest neighbor
    return resized_im


class TmpDirectory:
    def __init__(self, dirpath: str):
        self.cur_idx = 0
        self.dirpath = dirpath

    def __enter__(self):
        if os.path.exists(self.dirpath):
            for _ in os.listdir():
                raise FileExistsError('Temporary directory path must not contain existing files, please check if any '
                                      'file exists.')
        ensure_dir_exists(self.dirpath, remove_if_already_exists=True)
        return self

    def assign_tmpdir(self):
        """Return a path prefix that is guaranteed to be empty within the temporary directory

        Files and directories with this prefix can be created then
        """
        tmppath = f'{self.dirpath}/{self.cur_idx}_'
        self.cur_idx += 1
        return tmppath

    def __exit__(self, exc_type, exc_val, exc_tb):
        shutil.rmtree(self.dirpath)


def test():
    assert ImFileType.ftype_from_im_path('./some_root/hello.nii.gz') == ImFileType.FTYPE_NIB
    assert ImFileType.ftype_from_im_path('./some_root/hello_0001.JPG') == ImFileType.FTYPE_JPG
    assert ImFileType.ftype_from_im_path('./a.py') == ImFileType.FTYPE_UNKNOWN
    assert ImFileType.get_imid('./some_root/hello.png') == 'hello'
    assert (ImFileType.get_imid('./some_root/nibfile0001.nii.gz', is_stack=True, keep_dirpath=True) ==
            './some_root/nibfile')

    print('All tests successfully completed')



"""
DEPRECATED

This files define methods that are used to segment or post-processing a segmentation output
of image segmentation models. The get_counter() function in this folder selects a method
for segmenting images this way, and the returned method instance can be used to make predictions
on image cubes (W * W * W) to output a segmentation mask from 0 to N, where 0 is the background
class.
"""


import enum
from typing import Callable

import skimage
import skimage.morphology as morph
import scipy.ndimage as ndimage
import numpy as np
from abc import ABC, abstractmethod
import skimage.draw as draw
import matplotlib.pyplot as plt
from PIL import Image, ImageDraw, ImageFont


ZOOM_FACTOR = 4


def draw_ncells_text_on_im_cube(im_cube: np.ndarray, pos_list: np.ndarray, ncell_list: np.array) -> np.array:
    sz = (im_cube.shape[0], im_cube.shape[1] * ZOOM_FACTOR, im_cube.shape[2] * ZOOM_FACTOR) + (4,)
    out_cube = np.zeros(sz, dtype=np.uint8)
    out_cube[:, :, :, 3] = 255
    out_cube[:, :, :, 2] = np.array(ndimage.zoom(im_cube, (1, ZOOM_FACTOR, ZOOM_FACTOR), order=0) * 255, dtype=np.uint8)

    pos_map = {i: [] for i in range(im_cube.shape[0])}
    for i in range(len(pos_list)):
        pos_z = pos_list[i, 0].item()
        pos_map[pos_z].append(i)
    for i in range(im_cube.shape[0]):
        im = Image.fromarray(out_cube[i])
        # out_im = Image.new("RGBA", sz[1:], (255, 255, 255, 0))

        fnt = ImageFont.truetype("C:/Windows/Fonts/ariblk.ttf", 10)
        for j in pos_map[i]:
            pos = pos_list[j]
            context = ImageDraw.Draw(im)
            context.text((pos[2] * ZOOM_FACTOR, pos[1] * ZOOM_FACTOR),
                         f'{ncell_list[j].item():.1f}', font=fnt, fill=(255, 255, 255, 255))

        # out = Image.alpha_composite(base, out_im)
        # out.show()
        out_cube[i] = np.array(im, dtype=np.uint8)
    return out_cube


def grey_to_rgba(im_grey: np.array) -> np.array:
    out = np.stack((im_grey * 255,) * 4, axis=-1)
    out[..., :2] = 0  # blue

    return out


"""-------------------------Part 0: The interfaces----------------------------------------"""


class CountingMethod(enum.Enum):
    BLOBDOG = 0
    THRES_COUNT = 1
    THRES_BYSIZE = 2
    THRES_VOLUME = 3
    THRES_WATERSHED_COUNT = 4
    THRES_WATERSHED_BYSIZE = 5
    THRES_WATERSHED_VOLUME = 6
    CELLSEG3D_COUNT = 7
    CELLSEG3D_BYSIZE = 8
    CELLSEG3D_VOLUME = 9
    CELLSEG3D_WATERSHED_COUNT = 10
    CELLSEG3D_WATERSHED_BYSIZE = 11
    CELLSEG3D_WATERSHED_VOLUME = 12
    SUM_INTENSITY = 13


counting_method_dict = {item.name: item.value for item in CountingMethod}
counting_method_inverse_dict = {item.value: item.name for item in CountingMethod}


def get_counter(ty: int, get_mask: Callable):
    if ty == CountingMethod.SUM_INTENSITY:
        counter = Count_SumIntensity(130., .4)
    elif 1 <= ty <= 12:
        if ty <= 6:
            cellseg3d = False
            s1 = Mask_Thres(.45)
        else:
            cellseg3d = True
            s1 = Mask_CellSeg3D(get_mask)

        use_watershed = (ty - 1) % 6 >= 3
        if (ty - 1) % 3 == 2:
            # VOLUME ONLY
            if cellseg3d:
                if use_watershed:
                    ncell_from_inst = NCellFromInst_BySize(0., 212.)
                else:
                    ncell_from_inst = NCellFromInst_BySize(0., 220.)
            else:
                if use_watershed:
                    ncell_from_inst = NCellFromInst_BySize(0., 286.)
                else:
                    ncell_from_inst = NCellFromInst_BySize(0., 222.)
        elif (ty - 1) % 3 == 1:
            # BYSIZE
            if cellseg3d:
                if use_watershed:
                    ncell_from_inst = NCellFromInst_BySize(25., 152.)
                else:
                    ncell_from_inst = NCellFromInst_BySize(26., 140.)
            else:
                ncell_from_inst = NCellFromInst_BySize(40., 240.)
        else:
            # COUNT ONLY
            ncell_from_inst = NCellFromInst_BySize(1e10, 1e10)

        if use_watershed:
            s2 = CountFromMask_Watershed(ncell_from_inst,
                                         size_thres=60.,
                                         dist_thres=1.,
                                         rst=None,
                                         size_thres2=100.,
                                         dist_thres2=1.5,
                                         rst2=60.)
        else:
            s2 = CountFromMask_Direct(ncell_from_inst)

        counter = Count_TwoStage(s1, s2)
    else:
        counter = Count_BlobDog(max_sigma=2, threshold=0.1)
    return counter


class CountFromIntensityImage(ABC):
    @abstractmethod
    def count(self, im_cube: np.array) -> float:
        """
        Return the number of cells from image; The number should be counted one for each cell
        within the cube and 0.5 for each cell neighboring the boundaries (The counting of cells
        neighboring two or more walls of the cube area is 1/(num_neighbor + 1))

        The image is an float32 intensity image within range 0-1

        The return is a float estimate of (average of) the number of cells in the image given the image
        """
        raise NotImplementedError()

    @abstractmethod
    def interpretable(self, im_cube: np.array) -> np.array:
        """
        This method should return an RGBA-channel overlay of the same size of the original image, intended for
        interpretation of the prediction result.
        """
        raise NotImplementedError()


class Count_SumIntensity(CountFromIntensityImage):
    def __init__(self, intensity_per_cell, thres=0.):
        """
        thres (float) - intensity below this is ignored; intensity above this is directly summed without change
        """
        self.intensity_per_cell = intensity_per_cell
        self.thres = thres
        assert intensity_per_cell > 0, f'{intensity_per_cell}'
        assert 0. <= thres <= 1., f'{thres}'

    def count(self, im_cube: np.array):
        mask = im_cube > self.thres
        ncells = (im_cube * mask).sum() / self.intensity_per_cell
        return ncells

    def interpretable(self, im_cube: np.array) -> np.array:
        out = im_cube * (im_cube > self.thres)
        return grey_to_rgba(out)


class Count_BlobDog(CountFromIntensityImage):
    def __init__(self, **args):
        self.args = args

    def count(self, im_cube: np.array) -> float:
        blobs_dog = skimage.feature.blob_dog(np.array(im_cube * 255, dtype=np.uint8), **self.args)
        width = im_cube.shape[0]
        assert width == im_cube.shape[1] and width == im_cube.shape[2]
        border_dists = np.abs((blobs_dog[:, :3] + 16) % 32 - 15.5)  # compute border distances in z, y, x directions
        mults = 1 / np.clip(3. - border_dists * .5, 1., 2.)
        count = np.prod(mults, axis=1).sum()
        return count

    def interpretable(self, im_cube: np.array) -> np.array:
        blobs_dog = skimage.feature.blob_dog(np.array(im_cube * 255, dtype=np.uint8), **self.args)
        blobs_dog[:, 3:] = blobs_dog[:, 3:] * np.sqrt(3)  # adjust each sigma to get radius (this is still in pixels)

        blobs_img = np.zeros(im_cube.shape + (4, ), dtype=np.float32)
        for i in range(blobs_dog.shape[0]):
            o = blobs_dog[i]
            iz, iy, ix, ir = int(o[0]), int(o[1]), int(o[2]), o[3]
            rr, cc = draw.circle_perimeter(iy, ix, int(max(1., ir / 2)))
            ylim, xlim = blobs_img.shape[1:3]
            inds = (rr >= ylim) + (cc >= xlim) + (rr < 0) + (cc < 0) == 0
            rr, cc = rr[inds], cc[inds]
            blobs_img[iz, rr, cc] = (0, 0, 255, 255)

        return blobs_img


class Count_TwoStage(CountFromIntensityImage):
    """
    First create a binary mask from input image, then use the input image and binary mask to
    get a cell num estimate
    """

    def __init__(self, s1: "MaskFromIntensityImage", s2: "CountFromMask"):
        self.s1 = s1
        self.s2 = s2

    def count(self, im_cube: np.array) -> float:
        mask = self.s1.mask(im_cube)
        ncells = self.s2.count(mask)
        return ncells

    def interpretable(self, im_cube: np.array) -> np.array:
        mask = self.s1.mask(im_cube)
        pos_list, ncells_list = self.s2.count_annotate(mask)
        out_im = draw_ncells_text_on_im_cube(mask, pos_list, ncells_list)
        return out_im


"""-------------------------Part 1: Mask from input image---------------------------------"""


class MaskFromIntensityImage(ABC):
    @abstractmethod
    def mask(self, im: np.array):
        """
        Create a binary mask from the input image
        """
        raise NotImplementedError()


class Mask_Thres(MaskFromIntensityImage):
    def __init__(self, thres):
        self.thres = thres

    def mask(self, im: np.array):
        mask = im > self.thres
        return mask


class Mask_CellSeg3D(MaskFromIntensityImage):
    def __init__(self, get_mask):
        self.get_mask = get_mask

    def mask(self, im: np.array):
        return self.get_mask(im)


"""-------------------------Part 2: cell count from mask---------------------------------"""


class CountFromMask(ABC):
    def __init__(self, ncell_from_inst: "NCellFromInst"):
        self.ncell_from_inst = ncell_from_inst

    def count(self, mask: np.array) -> float:
        lbl_im = self.inst(mask)
        return self.ncell_from_inst.count(lbl_im)

    @abstractmethod
    def inst(self, mask: np.array) -> np.array:
        raise NotImplementedError()

    def count_annotate(self, mask: np.array) -> tuple[np.array, np.array]:
        lbl_im = self.inst(mask)
        return self.ncell_from_inst.count_annotate(lbl_im)


class CountFromMask_Direct(CountFromMask):
    def inst(self, mask: np.array) -> np.array:
        lbl_im, nlbl = ndimage.label(mask)
        return lbl_im


class CountFromMask_Watershed(CountFromMask):
    def __init__(self, ncell_from_inst: "NCellFromInst", **args):
        super().__init__(ncell_from_inst)
        self.args = args

    def inst(self, mask: np.array) -> np.array:
        lbl_im = round_object_detection(mask, **self.args)
        return lbl_im


"""-------------------------Part 3: ncells from instance segmentation---------------------"""


class NCellFromInst(ABC):
    def count(self, inst: np.array) -> float:
        return self.count_annotate(inst)[1].sum().item()

    @abstractmethod
    def count_annotate(self, inst: np.array) -> tuple[np.array, np.array]:
        """
        For each contour in the mask, compute the location of its centroid, and
        assign a number to it indicating how many cells it is worth
        Params
            inst - The mask to count on
        Returns
            A tuple of two items, in the order below:
            - A int32 array of size N * 3, locations of the centroids, where N is the #contours in the image
            - An array of size N, how many cells each contour worth
        """
        raise NotImplementedError()


class NCellFromInst_BySize(NCellFromInst):
    """
    Estimate the number of cells in cell clumps by looking at the size of the mask over it;
    small cells are still treated as one cell, while big clumps are counted as m cells where m
    depends on the number of voxels in the clump
    """

    def __init__(self, size_thres, voxel_per_cell):
        self.size_thres = size_thres
        self.voxel_per_cell = voxel_per_cell

    def count_annotate(self, inst: np.array) -> tuple[np.array, np.array]:
        assert len(inst.shape) == 3, f'ERROR: Input segmentation mask must be 3d, got shape {inst.shape}'
        D, H, W = inst.shape
        nlbl = inst.max().item()
        mins = np.array([0, 0, 0], dtype=np.float32)
        maxs = np.array([D - 1, H - 1, W - 1], dtype=np.float32)

        pos_list = []
        ncells_list = []
        for i in range(1, nlbl + 1):
            inds = np.argwhere(inst == i)
            if inds.shape[0] < 5:
                # contour is too small, ignore it
                continue

            # get a 3-vector of number of bordering pixels in z, y, x dimensions
            nborders = np.sum((inds == mins[None, :]) + (inds == maxs[None, :]), axis=0)
            nborders = np.clip(nborders, 0, 1)
            nadd = 1 / (nborders.sum() + 1)
            if inds.shape[0] > self.size_thres:
                nadd += (inds.shape[0] - self.size_thres) / self.voxel_per_cell
            pos_list.append(inds.mean(axis=0))
            ncells_list.append(nadd)
        pos_list = np.array(pos_list, dtype=np.int32)
        ncells_list = np.array(ncells_list, dtype=np.float32)
        return pos_list, ncells_list


"""-------------------------Part 4: Implementations----------------------------------------"""


def instance_to_binary(seg_inst):
    """
    A convenient function to convert a [0, N] mask to binary mask while still somewhat preserving instance
    separation: Neighboring instances will have a thin dividing line drawn between them.
    """
    D, H, W = seg_inst.shape
    seg_bin = np.zeros((D - 1, H - 1, W - 1), dtype=np.uint8)
    chunk = seg_inst[:D - 1, :H - 1, :W - 1]
    for d in range(3):
        mins = [0, 0, 0]
        maxs = [D - 1, H - 1, W - 1]
        mins[d] += 1
        maxs[d] += 1
        chunk_other = seg_inst[mins[0]:maxs[0], mins[1]:maxs[1], mins[2]:maxs[2]]
        edge_detected = (chunk == chunk_other) + (chunk == 0) + (chunk_other == 0) == 0
        seg_bin = np.logical_or(seg_bin, edge_detected)
    seg_inst = np.copy(seg_inst)
    seg_inst[:D - 1, :H - 1, :W - 1] = np.logical_and(seg_inst[:D - 1, :H - 1, :W - 1], ~seg_bin)
    return seg_inst


def watershed(seg_bin, dist_thres=1., remove_smaller_than=None):
    """
    Run Watershed algorithm to perform instance segmentation. The result is a index labeled int64 mask
    :param seg_bin: The binary [0, 1] 3d mask to run watershed on to separate blobs into instances.
    :param dist_thres: Only pixels this much into the contours are retained; pixels on contours surface are removed
    :param remove_smaller_than: Contours smaller than this value are added as part of neighboring contours (once)
    :return: The instance segmented int64 mask from 0 to N, where N is number of objects
    """
    # reference: https://docs.opencv.org/4.x/d3/db4/tutorial_py_watershed.html
    # fp_width = 2
    # fp = [(np.ones((fp_width, 1, 1)), 1), (np.ones((1, fp_width, 1)), 1), (np.ones((1, 1, fp_width)), 1)]
    # sure_bg = morph.binary_dilation(seg_bin, fp)
    sure_bg = seg_bin
    # sure_fg = morph.binary_erosion(seg_bin, fp)
    dist_transform = ndimage.distance_transform_edt(seg_bin)
    sure_fg = dist_transform >= dist_thres
    unknown = sure_bg ^ sure_fg
    lbl_im = morph.label(sure_fg, connectivity=1)

    # so that sure_bg is 1 and unknown region is 0
    lbl_im += 1
    lbl_im[unknown == 1] = 0
    result = skimage.segmentation.watershed(-dist_transform, lbl_im, connectivity=1)
    result -= result > 0  # we don't need to mark sure_bg as 1, make all marker id smaller by 1

    if remove_smaller_than is not None:
        # watershed again over the small object regions
        small_mask, big_mask = split_labeled_objects(result, remove_smaller_than, connectivity=1)
        lbl_im[small_mask] = 0
        result = skimage.segmentation.watershed(-dist_transform, lbl_im, connectivity=1)

        # at this point, there may be unfilled space, which are rare cases where lots of small objects make up a
        # larger connected part; these space should be cells but the individual objects do not meet size requirement
        unfilled = morph.label(result == 0, connectivity=1)
        result += (unfilled != 0) * (unfilled + result.max())

        result -= result > 0  # we don't need to mark sure_bg as 1, make all marker id smaller by 1

    return result


def split_labeled_objects(lbl_im, size_thres, connectivity=1):
    component_sizes = np.bincount(lbl_im.ravel())
    small_inds = component_sizes < size_thres
    small_inds[0] = False
    small_mask = small_inds[lbl_im]
    big_inds = component_sizes >= size_thres
    big_inds[0] = False
    big_mask = big_inds[lbl_im]
    return small_mask, big_mask


def split_objects(seg, size_thres, connectivity=1):
    lbl_im = morph.label(seg, connectivity=connectivity)
    return split_labeled_objects(lbl_im, size_thres, connectivity)


def round_object_detection(seg, size_thres, dist_thres, rst, size_thres2, dist_thres2, rst2):
    """
    detect round objects in a 3d image
    :param seg: the binary mask where we want to detect on
    :param size_thres: the size threshold below which size, contours are all kept as is
    :param dist_thres: threshold for seeding in the watershed algorithm
    :return: same shape image labeled 0, 1, ..., n, where 0 is background and 1...n are detected objects
    """
    # objects too small cannot be connected. We use this property to first find small objects that must be single cells
    small_mask, big_mask = split_objects(seg, size_thres, connectivity=1)
    big_mask, big_mask2 = split_objects(big_mask, size_thres2, connectivity=1)

    # big objects may be a single large cell or overlapping small cells, run watershed to separate overlapping cells
    lbl_im = watershed(big_mask, dist_thres=dist_thres, remove_smaller_than=rst)
    lbl_im2 = watershed(big_mask2, dist_thres=dist_thres2, remove_smaller_than=rst2)

    # finally, combine the two set of cells
    small_labeled = morph.label(small_mask, connectivity=1)
    lbl_im += (small_labeled != 0) * (small_labeled + lbl_im.max())
    lbl_im += (lbl_im2 != 0) * (lbl_im2 + lbl_im.max())
    # lbl_im = (lbl_im2 > 0) * 1 + (lbl_im > 0) * 2 + small_mask * 3

    return lbl_im


"""-------------------------Part 5: Statistical Analysis----------------------------------"""


def Stats_MAE(counted, gt):
    """
    Params
        counted (list) - the counted cells in each image
        gt (Iterable[float]) - the ground truth number of cells in each image
    Returns
        the mean absolute difference between counted and gt
    """
    return np.abs(np.array(counted, dtype=np.float32) - np.array(gt, dtype=np.float32)).mean().item()


def Stats_ShowScatterPairComparisons(counted: np.array, gt) -> None:
    """
    counted (a NCounter * NImages np.array) - the counted cells in each image, by each counter
    gt (Iterable[float]) - the ground truth number of cells in each image
    """
    gt_arr = np.array(gt, dtype=np.float32)
    fig, axes = plt.subplots(3, 6, figsize=(24, 12), sharex=True, sharey=True)
    for i in range(counted.shape[0]):
        X, Y = gt_arr, counted[i]

        iax, jax = i // 6, i % 6
        ax = axes[iax, jax]
        ax.set_box_aspect(1)
        ax.set_title(counting_method_inverse_dict[i])
        ax.scatter(X, Y)

    plt.show()
